"""
Collection of helper functions and classes related to the creation and processing of L{docutils} nodes.
"""
from __future__ import annotations

from typing import Iterable, Iterator, Optional, TypeVar, cast, TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Literal

import optparse

from docutils import nodes, utils, frontend, __version_info__ as docutils_version_info
from docutils.transforms import parts

__docformat__ = 'epytext en'

_DEFAULT_DOCUTILS_SETTINGS: Optional[optparse.Values] = None

def new_document(source: Literal['docstring', 'code'], settings: Optional[optparse.Values] = None) -> nodes.document:
    """
    Create a new L{nodes.document} using the provided settings or cached default settings.

    @returns: L{nodes.document} with a C{source} attribute that matches the provided source.
    """
    global _DEFAULT_DOCUTILS_SETTINGS
    # If we have docutils >= 0.19 we use get_default_settings to calculate and cache
    # the default settings. Otherwise we let new_document figure it out.
    if settings is None and docutils_version_info >= (0,19):
        if _DEFAULT_DOCUTILS_SETTINGS is None:
            _DEFAULT_DOCUTILS_SETTINGS = frontend.get_default_settings()

        settings = _DEFAULT_DOCUTILS_SETTINGS

    return utils.new_document(source, settings)

def _set_nodes_parent(nodes: Iterable[nodes.Node], parent: nodes.Element) -> Iterator[nodes.Node]:
    """
    Set the L{nodes.Node.parent} attribute of the C{nodes} to the defined C{parent}. 
    
    @returns: An iterator containing the modified nodes.
    """
    for node in nodes:
        node.parent = parent
        yield node

TNode = TypeVar('TNode', bound=nodes.Node)
def set_node_attributes(node: TNode, 
                        document: Optional[nodes.document] = None, 
                        lineno: Optional[int] = None, 
                        children: Optional[Iterable[nodes.Node]] = None) -> TNode:
    """
    Set the attributes of a Node and return the modified node.
    This is required to manually construct a docutils document that is consistent.

    @param node: A node to edit.
    @param document: The L{nodes.Node.document} attribute.
    @param lineno: The L{nodes.Node.line} attribute.
    @param children: The L{nodes.Element.children} attribute. Special care is taken 
        to appropriately set the L{nodes.Node.parent} attribute on the child nodes. 
    """
    if lineno is not None:
        node.line = lineno
    
    if document:
        node.document = document

    if children:
        assert isinstance(node, nodes.Element), (f'Cannot set the children on Text node: "{node.astext()}". '
                                                 f'Children: {children}')
        node.extend(_set_nodes_parent(children, node))

    return node

def build_table_of_content(node: nodes.Element, depth: int, level: int = 0) -> nodes.Element | None:
    """
    Simplified from docutils Contents transform. 

    All section nodes MUST have set attribute 'ids' to a list of strings.
    """

    def _copy_and_filter(node: nodes.Element) -> nodes.Element:
        """Return a copy of a title, with references, images, etc. removed."""
        if (doc:=node.document) is None:
            raise AssertionError(f'missing document attribute on {node}')
        visitor = parts.ContentsFilter(doc)
        node.walkabout(visitor)
        #                                 the stubs are currently imcomplete, 2024.
        return visitor.get_entry_text() # type:ignore

    level += 1
    sections = [sect for sect in node if isinstance(sect, nodes.section)]
    entries = []
    if (doc:=node.document) is None:
        raise AssertionError(f'missing document attribute on {node}')
    
    for section in sections:
        title = cast(nodes.Element, section[0]) # the first element of a section is the header.
        entrytext = _copy_and_filter(title)
        reference = nodes.reference('', '', refid=section['ids'][0],
                                    *entrytext)
        ref_id = doc.set_id(reference, suggested_prefix='toc-entry')
        entry = nodes.paragraph('', '', reference)
        item = nodes.list_item('', entry)
        if title.next_node(nodes.reference) is None:
            title['refid'] = ref_id
        if level < depth:
            subsects = build_table_of_content(section, depth=depth, level=level)
            item += subsects or []
        entries.append(item)
    if entries:
        contents = nodes.bullet_list('', *entries)
        return contents
    else:
        return None

def get_lineno(node: nodes.Element) -> int:
    """
    Get the 0-based line number for a docutils `nodes.title_reference`.

    Walk up the tree hierarchy until we find an element with a line number, then
    counts the number of newlines until the reference element is found.
    """
    # Fixes https://github.com/twisted/pydoctor/issues/237
        
    def get_first_parent_lineno(_node: nodes.Element | None) -> int:
        if _node is None:
            return 0
        
        if _node.line:
            # This line points to the start of the containing node
            # Here we are removing 1 to the result because ParseError class is zero-based
            # while docutils line attribute is 1-based.
            line:int = _node.line-1
            # Let's figure out how many newlines we need to add to this number 
            # to get the right line number.
            parent_rawsource: Optional[str] = _node.rawsource or None
            node_rawsource: Optional[str] = node.rawsource or None

            if parent_rawsource is not None and \
               node_rawsource is not None:
                if node_rawsource in parent_rawsource:
                    node_index = parent_rawsource.index(node_rawsource)
                    # Add the required number of newlines to the result
                    line += parent_rawsource[:node_index].count('\n')
        else:
            line = get_first_parent_lineno(_node.parent)
        return line

    if node.line:
        # If the line is explicitely set, assume it's zero-based
        line = node.line
        # If docutils suddenly starts populating the line attribute for
        # title_reference node, all RST xref warnings will off by 1 :/

    else:
        line = get_first_parent_lineno(node.parent)
    
    return line

def text_node(text: str, klass: str | None = None) -> nodes.inline:
    """
    Create an inline node with the given text and class.
    """
    return set_node_attributes(
        nodes.inline('', '', classes=[klass] if klass else []), 
        children=[nodes.Text(text)],
    )

# additional docutils nodes: 

class wbr(nodes.inline):
    """
    Word break opportunity.
    """
    def __init__(self) -> None:
        super().__init__('', '')

class obj_reference(nodes.title_reference):
    """
    A reference to a documentable object.
    """

class code(nodes.inline):
    """
    Like a inline[class='literal'], but more elegant.
    """



# Part of the sphinx.pycode.parser module.
# Copyright 2007-2020 by the Sphinx team, see AUTHORS.
# BSD, see LICENSE for details.
from __future__ import annotations

import ast
import inspect
import re
from token import DEDENT, INDENT, NAME, NEWLINE, NUMBER, OP, STRING
from tokenize import COMMENT, generate_tokens, tok_name
from typing import Any, Sequence

class Token:
    """
    Better token wrapper for tokenize module.
    
    Example usage:
    >>> from token import NAME
    >>> t = Token(NAME, 'foo', (1, 0), (1, 3), 'foo = 1')
    >>> t.kind
    1
    >>> t.value
    'foo'
    >>> t.start
    (1, 0)
    >>> t.end
    (1, 3)
    >>> t.source
    'foo = 1'
    >>> t == NAME
    True
    >>> t == 'foo'
    True
    >>> t == ['foo']  # Not matching kind+value, returns False
    False
    >>> t == None
    False
    >>> t.match(NAME, 'foo')
    True
    >>> repr(t)
    "<Token kind='NAME' value='foo'>"
    >>> t == 3.14
    False
    """

    def __init__(self, kind: int, 
                 value: str, 
                 start: tuple[int, int], 
                 end: tuple[int, int],
                 source: str) -> None:
        self.kind = kind
        self.value = value
        self.start = start
        self.end = end
        self.source = source #: Source line

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, Token):
            return (self.kind == other.kind and
                self.value == other.value)
        elif isinstance(other, int):
            return self.kind == other
        elif isinstance(other, str):
            return self.value == other
        elif isinstance(other, list):
            return [self.kind, self.value] == other
        elif isinstance(other, tuple):
            return (self.kind, self.value) == other
        return NotImplemented

    def match(self, *conditions: Any) -> bool:
        return any(self == candidate for candidate in conditions)

    def __repr__(self) -> str:
        return '<Token kind=%r value=%r>' % (tok_name[self.kind],
                                             self.value.strip())

class TokenProcessor:
    r"""
    Processes tokens from source code lines.

    Example usage:
    >>> from token import NAME, OP, NUMBER, STRING
    >>> src = ["foo = 123 # comment\n"]
    >>> tp = TokenProcessor(src)
    >>> isinstance(tp.buffers, list)
    True

    # get_line returns the line at given lineno (1-based)
    >>> tp.get_line(1)
    'foo = 123 # comment\n'

    # fetch_token yields tokens one by one
    >>> t1 = tp.fetch_token()
    >>> t1.kind == NAME
    True
    >>> t2 = tp.fetch_token()
    >>> t2.kind == OP
    True
    >>> t3 = tp.fetch_token()
    >>> t3.kind == NUMBER
    True

    # previous and current attributes
    >>> tp.previous == t2
    True
    >>> tp.current == t3
    True

    # fetch_until collects tokens until matching condition
    >>> tp2 = TokenProcessor(["foo = (1 + 2)\n"])
    >>> tokens = tp2.fetch_until([OP, ')'])
    >>> any(t.value == ')' for t in tokens)
    True
    """

    def __init__(self, buffers: Sequence[str]) -> None:
        lines = iter(buffers)
        self.buffers = buffers
        self.tokens = generate_tokens(lambda: next(lines))
        self.current: Token | None = None
        self.previous: Token | None = None

    def get_line(self, lineno: int) -> str:
        """Returns specified line."""
        return self.buffers[lineno - 1]

    def fetch_token(self) -> Token | None:
        """Fetch a next token from source code.

        Returns ``None`` if sequence finished.
        """
        try:
            self.previous = self.current
            self.current = Token(*next(self.tokens))
        except StopIteration:
            self.current = None

        return self.current

    def fetch_until(self, condition: Any) -> list[Token]:
        """Fetch tokens until specified token appeared.

        .. note:: This also handles parenthesis well.
        """
        tokens = []
        while self.fetch_token():
            assert self.current
            tokens.append(self.current)
            if self.current == condition:
                break
            elif self.current == [OP, '(']:
                tokens += self.fetch_until([OP, ')'])
            elif self.current == [OP, '{']:
                tokens += self.fetch_until([OP, '}'])
            elif self.current == [OP, '[']:
                tokens += self.fetch_until([OP, ']'])

        return tokens


class AfterCommentParser(TokenProcessor):
    """Python source code parser to pick up comment after assignment.

    This parser takes a python code starts with assignment statement,
    and returns the comments for variable if exists.
    """

    def __init__(self, lines: Sequence[str]) -> None:
        super().__init__(lines)
        self.comment: str | None = None 

    def fetch_rvalue(self) -> Sequence[Token]:
        """Fetch right-hand value of assignment."""
        tokens: list[Token] = []
        while self.fetch_token():
            assert self.current
            tokens.append(self.current)
            if self.current == [OP, '(']:
                tokens += self.fetch_until([OP, ')'])
            elif self.current == [OP, '{']:
                tokens += self.fetch_until([OP, '}'])
            elif self.current == [OP, '[']:
                tokens += self.fetch_until([OP, ']'])
            elif self.current == INDENT:
                tokens += self.fetch_until(DEDENT)
            elif self.current == [OP, ';']:
                break
            elif self.current.kind not in (OP, NAME, NUMBER, STRING):
                break

        return tokens

    def parse(self) -> None:
        """Parse the code and obtain comment after assignment."""
        # skip lvalue (or whole of AnnAssign)
        while (current:=self.fetch_token()) and not current.match([OP, '='], NEWLINE, COMMENT):
            assert self.current
            continue

        # skip rvalue (if exists)
        if self.current == [OP, '=']:
            self.fetch_rvalue()

        if self.current == COMMENT:
            assert self.current
            self.comment = self.current.value

comment_re = re.compile('^\\s*#: ?(.*)\r?\n?$')
indent_re = re.compile('^\\s*$')

def extract_doc_comment_after(node: ast.Assign | ast.AnnAssign, lines: Sequence[str]) ->  tuple[int, str] | None:
    """
    Support for doc comment as found in sphinx.

    @param node: the assignment node
    @param lines: the lines of the source code of the module, as generated by
        C{code.splitlines(keepends=True)}.
    @returns: A tuple linenumber, docstring or None if the assignment doesn't have a doc comment.
    """
    # check doc comments after assignment
    current_line = lines[node.lineno - 1]
    parser = AfterCommentParser([current_line[node.col_offset:], *lines[node.lineno:]])
    parser.parse()
    if parser.comment and comment_re.match(parser.comment):
        docstring = comment_re.sub('\\1', parser.comment)
        return node.lineno, docstring

    return None

def extract_doc_comment_before(node: ast.Assign | ast.AnnAssign, lines: Sequence[str]) ->  tuple[int, str] | None:
    """
    Same as L{extract_doc_comment_after} but fetch the comment before the assignment.
    """
    # check doc comments before assignment
    comment_lines = []
    for i in range(node.lineno - 1):
        before_line = lines[node.lineno - 2 - i]
        if comment_re.match(before_line):
            comment_lines.append(comment_re.sub('\\1', before_line))
        else:
            break
    if comment_lines:
        docstring = inspect.cleandoc('\n'.join(reversed(comment_lines)))
        return node.lineno - len(comment_lines), docstring

    return None

# This was part of the sphinx.pycode.parser module.


"""
Render types from L{docutils.nodes.document} objects. 

This module provides yet another L{ParsedDocstring} subclass.
"""
from __future__ import annotations

from typing import Callable, Dict

from pydoctor.epydoc.markup import ParsedDocstring
from pydoctor.epydoc.markup._pyval_repr import PyvalColorizer
from pydoctor.napoleon.docstring import TokenType, ITokenizer, Tokenizer
from pydoctor.epydoc.docutils import new_document, set_node_attributes, code

from docutils import nodes

class NodeTokenizer(ITokenizer[nodes.document]):
    """
    A type tokenizer for annotation as docutils L{document <nodes.document>}.
    """

    def __init__(self, annotation: nodes.document, *, 
                 warns_on_unknown_tokens: bool) -> None:
        # build tokens and warnings
        self.warnings = warnings = [] # type: list[str]
        raw_tokens = Tokenizer.recombine_sets(self.tokenize_document(annotation, warnings))
        self.tokens = Tokenizer.build(raw_tokens, warnings, warns_on_unknown_tokens)
    
    @staticmethod
    def tokenize_document(spec: nodes.document, warnings: list[str]) -> list[str | nodes.Node]:
        def _warn_not_supported(n:nodes.Node) -> None:
            warnings.append("Unexpected element in type specification field: "
                                 f"element '{n.__class__.__name__}'. This value should "
                                 "only contain text or inline markup.")

        tokens: list[str | nodes.Node] = []
        # Determine if the content is nested inside a paragraph
        # this is generally the case, except for consolidated fields generate documents.
        if spec.children and isinstance(spec.children[0], nodes.paragraph):
            if len(spec.children)>1:
                _warn_not_supported(spec.children[1])
            children = spec.children[0].children
        else:
            children = spec.children
        
        for child in children:
            if isinstance(child, nodes.Text):
                # Tokenize the Text node with the same method TypeDocstring uses.
                tokens.extend(Tokenizer.tokenize_str(child.astext()))
            elif isinstance(child, nodes.Inline):
                tokens.append(child)
            else:
                _warn_not_supported(child)
        
        return tokens


class ParsedTypeDocstring(ParsedDocstring):
    """
    Add L{ParsedDocstring} interface on top of L{TypeDocstring} and 
    allow to parse types from L{nodes.Node} objects, 
    providing the C{--process-types} option.
    """

    FIELDS = ('type', 'rtype', 'ytype', 'returntype', 'yieldtype')

    def __init__(self, annotation: nodes.document, 
                 warns_on_unknown_tokens: bool = False, 
                 lineno: int = 0) -> None:
        super().__init__(fields=())

        tokenizer = NodeTokenizer(annotation, 
                              warns_on_unknown_tokens=warns_on_unknown_tokens)
        self._tokens = tokenizer.tokens
        self.warnings = tokenizer.warnings
        self._lineno = lineno
        self._document = self._parse_tokens()

    @property
    def has_body(self) -> bool:
        return len(self._tokens)>0

    def to_node(self) -> nodes.document:
        return self._document

    _converters: Dict[TokenType, Callable[[str, int], nodes.Node]] = {
            TokenType.LITERAL: lambda _token, _: nodes.inline(
                                         # we're re-using the STRING_TAG css 
                                         # class for the whole literal token, it's the
                                         # best approximation we have for now. 
                _token, _token, classes=[PyvalColorizer.STRING_TAG]),
            TokenType.CONTROL: lambda _token, _: nodes.emphasis(_token, _token),
            TokenType.OBJ: lambda _token, lineno: set_node_attributes(
                nodes.title_reference(_token, _token), lineno=lineno),
        }

    def _parse_tokens(self) -> nodes.document:
        """
        Convert type to docutils document object.
        """

        document = new_document('code')

        converters = self._converters
        lineno = self._lineno

        elements: list[nodes.Node] = []
        default = lambda _token, _: nodes.Text(_token)

        for _tok in self._tokens:
            token, type_ = _tok.value, _tok.type
            assert token is not None
            converted_token: nodes.Node
            
            if type_ is TokenType.ANY:
                assert isinstance(token, nodes.Node)
                converted_token = token
            else:
                assert isinstance(token, str)
                converted_token = converters.get(type_, default)(token, lineno)

            elements.append(set_node_attributes(converted_token, 
                                                    document=document))

        return set_node_attributes(document, children=[
            set_node_attributes(code('', ''), 
                                children=elements, 
                                document=document, 
                                lineno=lineno+1)])
                                # the +1 here is coping with the fact that
                                # Field.lineno are 0-based but the docutils tree 
                                # is supposed to be 1-based




"""Badly named module that contains the driving code for the rendering."""
from __future__ import annotations

import itertools
from pathlib import Path
import shutil
from typing import IO, Iterable, Type, TYPE_CHECKING

from pydoctor import model
from pydoctor.extensions import zopeinterface
from pydoctor.templatewriter import (
    DOCTYPE, pages, summary, search, TemplateLookup, IWriter, StaticTemplate
)

from twisted.python.failure import Failure
from twisted.web.template import flattenString

if TYPE_CHECKING:
    from twisted.web.template import Flattenable


def flattenToFile(fobj: IO[bytes], elem: "Flattenable") -> None:
    """
    This method writes a page to a HTML file.
    @raises Exception: If the L{twisted.web.template.flatten} call fails.
    """
    fobj.write(DOCTYPE)
    err = None
    def e(r: Failure) -> None:
        nonlocal err
        err = r.value
    flattenString(None, elem).addCallback(fobj.write).addErrback(e)
    if err:
        raise err


class TemplateWriter(IWriter):
    """
    HTML templates writer.
    """

    @classmethod
    def __subclasshook__(cls, subclass: Type[object]) -> bool:
        for name in dir(cls):
            if not name.startswith('_'):
                if not hasattr(subclass, name):
                    return False
        return True

    def __init__(self, build_directory: Path, template_lookup: TemplateLookup):
        """
        @arg build_directory: Build directory.
        @arg template_lookup: L{TemplateLookup} object.
        """
        self.build_directory = build_directory
        """Build directory"""

        self.template_lookup: TemplateLookup = template_lookup
        """Writer's L{TemplateLookup} object"""

        self.written_pages: int = 0
        self.total_pages: int = 0
        self.dry_run: bool = False
        

    def prepOutputDirectory(self) -> None:
        """
        Write static CSS and JS files to build directory.
        """
        self.build_directory.mkdir(exist_ok=True, parents=True)
        for template in self.template_lookup.templates:
            if isinstance(template, StaticTemplate):
                template.write(self.build_directory)

    def writeIndividualFiles(self, obs: Iterable[model.Documentable]) -> None:
        """
        Iterate through C{obs} and call L{_writeDocsFor} method for each L{Documentable}.
        """
        self.dry_run = True
        for ob in obs:
            self._writeDocsFor(ob)
        self.dry_run = False
        for ob in obs:
            self._writeDocsFor(ob)

    def writeSummaryPages(self, system: model.System) -> None:
        import time
        for pclass in itertools.chain(summary.summaryPages(system), search.searchpages):
            system.msg('html', 'starting ' + pclass.__name__ + ' ...', nonl=True)
            T = time.time()
            page = pclass(system=system, template_lookup=self.template_lookup)
            with self.build_directory.joinpath(pclass.filename).open('wb') as fobj:
                flattenToFile(fobj, page)
            system.msg('html', "took %fs"%(time.time() - T), wantsnl=False)
        
        # Generate the searchindex.json file
        system.msg('html', 'starting lunr search index ...', nonl=True)
        T = time.time()
        search.write_lunr_index(self.build_directory, system=system)
        system.msg('html', "took %fs"%(time.time() - T), wantsnl=False)
    
    def writeLinks(self, system: model.System) -> None:
        if len(system.root_names) == 1:
            # If there is just a single root module it is written to index.html to produce nicer URLs.
            # To not break old links we also create a link from the full module name to the index.html
            # file. This is also good for consistency: every module is accessible by <full module name>.html
            root_module_path = (self.build_directory / (list(system.root_names)[0] + '.html'))
            root_module_path.unlink(missing_ok=True) # introduced in Python 3.8
            
            try:
                if system.options.use_hardlinks:
                    # The use wants only harlinks, so simulate an OSError 
                    # to jump directly to the hardlink part.
                    raise OSError()
                root_module_path.symlink_to('index.html')
            except (OSError, NotImplementedError): # symlink is not implemented for windows on pypy :/ 
                hardlink_path = (self.build_directory / 'index.html')
                shutil.copy(hardlink_path, root_module_path)

    def _writeDocsFor(self, ob: model.Documentable) -> None:
        if not ob.isVisible:
            return
        if ob.documentation_location is model.DocLocation.OWN_PAGE:
            if self.dry_run:
                self.total_pages += 1
            else:
                with self.build_directory.joinpath(ob.url).open('wb') as fobj:
                    self._writeDocsForOne(ob, fobj)
        for o in ob.contents.values():
            self._writeDocsFor(o)

    def _writeDocsForOne(self, ob: model.Documentable, fobj: IO[bytes]) -> None:
        if not ob.isVisible:
            return
        pclass: Type[pages.CommonPage] = pages.CommonPage
        class_name = ob.__class__.__name__
        
        # Special case the zope interface custom renderer. 
        # TODO: Find a better way of handling renderer customizations and get rid of ZopeInterfaceClassPage completely.
        if class_name == 'Class' and isinstance(ob, zopeinterface.ZopeInterfaceClass):
            class_name = 'ZopeInterfaceClass'
        
        try:
            # This implementation relies on 'pages.commonpages' dict that ties
            # documentable class name (i.e. 'Class') with the
            # page class used for rendering: pages.ClassPage
            pclass = pages.commonpages[class_name]
        except KeyError:
            ob.system.msg(section="html", 
                # This is typically only reached in tests, when rendering Functions or Attributes with this method.
                msg=f"Could not find page class suitable to render object type: {class_name!r}, using CommonPage.", 
                once=True, thresh=-2)
        
        ob.system.msg('html', str(ob), thresh=1)
        page = pclass(ob=ob, template_lookup=self.template_lookup)
        self.written_pages += 1
        ob.system.progress('html', self.written_pages, self.total_pages, 'pages written')
        flattenToFile(fobj, page)



from typing import Any, List
from textwrap import dedent
from pydoctor.epydoc.markup import ParseError, get_parser_by_name
from pydoctor.test.epydoc.test_restructuredtext import prettify
from pydoctor.test.test_templatewriter import getHTMLOfAttribute
from pydoctor.test import NotFoundLinker, CapSys
from pydoctor.test.epydoc import parse_docstring
from pydoctor.test.test_epydoc2stan import docstring2html
from pydoctor.test.test_astbuilder import fromText
from pydoctor.stanutils import flatten
from pydoctor.epydoc.markup._types import ParsedTypeDocstring
import pydoctor.epydoc.markup
from pydoctor import model


def doc2html(doc: str, markup: str, processtypes: bool = False) -> str:
    return ''.join(prettify(flatten(parse_docstring(doc, markup, processtypes).to_stan(NotFoundLinker()))).splitlines())

def test_types_to_node_no_markup() -> None:
    cases = [   
            'rtype: list of int or float or None', 
            "rtype: {'F', 'C', 'N'}, default 'N'",
            "rtype: DataFrame, optional",
            "rtype: List[str] or list(bytes), optional",]

    for s in cases:
        assert doc2html(':'+s, 'restructuredtext', False) == doc2html('@'+s, 'epytext')
        assert doc2html(':'+s, 'restructuredtext', True) == doc2html('@'+s, 'epytext')

def test_to_node_markup() -> None:
    
    cases = [  ('L{me}', '`me`'),
            ('B{No!}', '**No!**'),
            ('I{here}', '*here*'),
            ('L{complicated string} or L{strIO <twisted.python.compat.NativeStringIO>}', '`complicated string` or `strIO <twisted.python.compat.NativeStringIO>`')
            ]

    for epystr, rststr in cases:
        assert doc2html(rststr, 'restructuredtext') == doc2html(epystr, 'epytext')

def typespec2htmlvianode(s: str, markup: str) -> str:
    err: List[ParseError] = []
    parsed_doc = get_parser_by_name(markup)(s, err)
    assert not err
    ann = ParsedTypeDocstring(parsed_doc.to_node(), warns_on_unknown_tokens=True)
    html = flatten(ann.to_stan(NotFoundLinker()))
    assert not ann.warnings
    return html

def test_parsed_type(subtests: Any) -> None:
    
    parsed_type_cases = [
        ('list of int or float or None', 
        '<code><a>list</a> of <a>int</a> or <a>float</a> or <a>None</a></code>'),

        ("{'F', 'C', 'N'}, default 'N'",
        """<code><span class="rst-variable-string">{'F', 'C', 'N'}</span>, <em>default</em> <span class="rst-variable-string">'N'</span></code>"""),

        ("DataFrame, optional",
        """<code><a>DataFrame</a>, <em>optional</em></code>"""),

        ("List[str] or list(bytes), optional", 
        """<code><a>List</a>[<a>str</a>] or <a>list</a>(<a>bytes</a>), <em>optional</em></code>"""),

        (('`complicated string` or `strIO <twisted.python.compat.NativeStringIO>`', 'L{complicated string} or L{strIO <twisted.python.compat.NativeStringIO>}'),
        '<code><a>complicated string</a> or <a>strIO</a></code>'),
    ]

    for string, excepted_html in parsed_type_cases:
        rst_string = ''
        epy_string = ''

        if isinstance(string, tuple):
            rst_string, epy_string = string
        elif isinstance(string, str):
            rst_string = epy_string = string

        with subtests.test('parse type', rst=rst_string, epy=epy_string):
        
            assert typespec2htmlvianode(rst_string, 'restructuredtext') == excepted_html            
            assert typespec2htmlvianode(epy_string, 'epytext') == excepted_html

def test_processtypes(capsys: CapSys) -> None:
    """
    Currently, numpy and google type parsing happens both at the string level with L{pydoctor.napoleon.docstring.TypeDocstring}
    and at the docutils nodes L{ParsedTypeDocstring} for type fields (``type`` and ``rtype``).
    """

    cases = [
        (
            (   
                """
                @param arg: A param.
                @type arg: list of int or float or None
                """,

                """
                :param arg: A param.
                :type arg: list of int or float or None
                """,

                """
                Args:
                    arg (list of int or float or None): A param.
                """,

                """
                Args
                ----
                arg: list of int or float or None
                    A param.
                """,
            ), 

                ("list of int or float or None", 
                '<code><a>list</a> of <a>int</a> or <a>float</a> or <a>None</a></code>')

        ),

        (
            (   
                """
                @param arg: A param.
                @type arg: L{complicated string} or L{strIO <twisted.python.compat.NativeStringIO>}, optional
                """,

                """
                :param arg: A param.
                :type arg: `complicated string` or `strIO <twisted.python.compat.NativeStringIO>`, optional
                """,

                """
                Args:
                    arg (`complicated string` or `strIO <twisted.python.compat.NativeStringIO>`, optional): A param.
                """,

                """
                Args
                ----
                arg: `complicated string` or `strIO <twisted.python.compat.NativeStringIO>`, optional
                    A param.
                """,
            ), 

                ("<code><a>complicated string</a></code> or <code><a>strIO</a></code>, optional", 
                '<code><a>complicated string</a> or <a>strIO</a>, <em>optional</em></code>')

        ),

    ]

    for strings, excepted_html in cases:
        epy_string, rst_string, goo_string, numpy_string = strings

        excepted_html_no_process_types, excepted_html_type_processed = excepted_html

        assert flatten(parse_docstring(epy_string, 'epytext').fields[-1].body().to_stan(NotFoundLinker())) == excepted_html_no_process_types
        assert flatten(parse_docstring(rst_string, 'restructuredtext').fields[-1].body().to_stan(NotFoundLinker())) == excepted_html_no_process_types

        assert flatten(parse_docstring(dedent(goo_string), 'google').fields[-1].body().to_stan(NotFoundLinker())) == excepted_html_type_processed
        assert flatten(parse_docstring(dedent(numpy_string), 'numpy').fields[-1].body().to_stan(NotFoundLinker())) == excepted_html_type_processed

        assert flatten(parse_docstring(epy_string, 'epytext', processtypes=True).fields[-1].body().to_stan(NotFoundLinker())) == excepted_html_type_processed
        assert flatten(parse_docstring(rst_string, 'restructuredtext', processtypes=True).fields[-1].body().to_stan(NotFoundLinker())) == excepted_html_type_processed

def test_processtypes_more() -> None:
    # Using numpy style-only because it suffice.
    cases = [
        ("""
              Yields
              ------
              working: bool
                  Whether it's working.
              not_working: bool
                  Whether it's not working.
              """, 
              """<ul class="rst-simple">
<li><strong>working</strong>: <code><a>bool</a></code> - Whether it's working.</li>
<li><strong>not_working</strong>: <code><a>bool</a></code> - Whether it's not working.</li>
</ul>"""), 

              ("""
               Returns
               -------
               name: str
                  the name description.
               content: str
                  the content description.
               """, 
               """<ul class="rst-simple">
<li><strong>name</strong>: <code><a>str</a></code> - the name description.</li>
<li><strong>content</strong>: <code><a>str</a></code> - the content description.</li>
</ul>"""),
              ]
    
    for string, excepted_html in cases:
        assert flatten(parse_docstring(dedent(string), 'numpy').fields[-1].body().to_stan(NotFoundLinker())).strip() == excepted_html

def test_processtypes_with_system(capsys: CapSys) -> None:
    system = model.System()
    system.options.processtypes = True
    
    mod = fromText('''
    a = None
    """
    Variable documented by inline docstring.
    @type: list of int or float or None
    """
    ''', modname='test', system=system)
    a = mod.contents['a']
    
    docstring2html(a)
    assert isinstance(a.parsed_type, ParsedTypeDocstring)
    fmt = flatten(a.parsed_type.to_stan(NotFoundLinker()))

    captured = capsys.readouterr().out
    assert not captured

    assert '<code><a>list</a> of <a>int</a> or <a>float</a> or <a>None</a></code>' == fmt
    

def test_processtypes_corner_cases(capsys: CapSys, subtests: Any) -> None:
    """
    The corner cases does not trigger any warnings because they are still valid types.
    
    Warnings should be triggered in L{pydoctor.napoleon.docstring.TypeDocstring._trigger_warnings}, 
    we should be careful with triggering warnings because whether the type spec triggers warnings is used
    to check is a string is a valid type or not.  
    """
    def _process(typestr: str, fails:bool=False, docformat:str='both') -> str:
        if docformat == 'both':
            str1 = _process(typestr, fails, 'epytext')
            str2 = _process(typestr, fails, 'restructuredtext')
            assert str1 == str2
            return str1
        
        system = model.System()
        system.options.processtypes = True
        mod = fromText(f'''
        __docformat__ = '{docformat}'
        a = None
        """
        {'@' if docformat == 'epytext' else ':'}type: {typestr}
        """
        ''', modname='test', system=system)
        a = mod.contents['a']
        docstring2html(a)

        assert isinstance(a.parsed_type, ParsedTypeDocstring)
        fmt = flatten(a.parsed_type.to_stan(NotFoundLinker()))
        assert fmt.startswith(b:='<code>')
        assert fmt.endswith(e:='</code>')
        fmt = fmt[len(b):-(len(e))]
        
        if not fails:
            captured = capsys.readouterr().out
            assert not captured

        return fmt

    def process(input:str, expected:str, fails:bool=False, docformat:str='both') -> None:
        # both is for epytext and restructuredtext
        with subtests.test(msg="processtypes", input=input):
            actual = _process(input, fails=fails, docformat=docformat)
            assert actual == expected

    process('default[str]',                       "<em>default</em>[<a>str</a>]")
    process('[str]',                              "[<a>str</a>]")
    process('[,]',                                "[, ]")
    process('[[]]',                               "[[]]")
    process(', [str]',                            ", [<a>str</a>]")
    process(' of [str]',                          "of [<a>str</a>]")
    process(' or [str]',                          "or [<a>str</a>]")
    process(': [str]',                            ': [<a>str</a>]')
    process("'hello'[str]",                      "<span class=\"rst-variable-string\">'hello'</span>[<a>str</a>]")
    process('"hello"[str]',                       "<span class=\"rst-variable-string\">\"hello\"</span>[<a>str</a>]")
    process('["hello" or str, default: 2]',       """[<span class="rst-variable-string">"hello"</span> or <a>str</a>, <em>default</em>: <span class="rst-variable-string">2</span>]""")
    
    process('`hello`[str]',                       "`hello`[<a>str</a>]", fails=True, docformat='restructuredtext')
    process('`hello <https://github.com>`_[str]', """`hello &lt;<a class="rst-external rst-reference" href="https://github.com" target="_top">https://github.com</a>&gt;`_[<a>str</a>]""", fails=True, docformat='restructuredtext')
    process('**hello**[str]',                     "**hello**[<a>str</a>]", fails=True, docformat='restructuredtext')
   
    # HTML ids for problematic elements changed in docutils 0.18.0, and again in 0.19.0, so we're not testing for the exact content anymore.
    with subtests.test(msg="processtypes", input='Union[`hello <>`_[str]]'):
        problematic = _process('Union[`hello <>`_[str]]', fails=True, docformat='restructuredtext')
        assert "`hello &lt;&gt;`_" in problematic
        assert "<a>str</a>" in problematic
 
def test_processtypes_warning_unexpected_element(capsys: CapSys) -> None:
    

    epy_string = """
    @param arg: A param.
    @type arg: L{complicated string} or 
        L{strIO <twisted.python.compat.NativeStringIO>}, optional
        
        >>> print('example')
    """

    rst_string = """
    :param arg: A param.
    :type arg: `complicated string` or 
        `strIO <twisted.python.compat.NativeStringIO>`, optional
        
        >>> print('example')
    """

    expected = """<code><a>complicated string</a> or <a>strIO</a>, <em>optional</em></code>"""
    
    # Test epytext
    epy_errors: List[ParseError] = []
    epy_parsed = pydoctor.epydoc.markup.processtypes(get_parser_by_name('epytext'))(epy_string, epy_errors)

    assert len(epy_errors)==1
    assert "Unexpected element in type specification field: element 'doctest_block'" in epy_errors.pop().descr()

    assert flatten(epy_parsed.fields[-1].body().to_stan(NotFoundLinker())).replace('\n', '') == expected
    
    # Test restructuredtext
    rst_errors: List[ParseError] = []
    rst_parsed = pydoctor.epydoc.markup.processtypes(get_parser_by_name('restructuredtext'))(rst_string, rst_errors)

    assert len(rst_errors)==1
    assert "Unexpected element in type specification field: element 'doctest_block'" in rst_errors.pop().descr()

    assert flatten(rst_parsed.fields[-1].body().to_stan(NotFoundLinker())).replace('\n', ' ') == expected

def test_napoleon_types_warnings(capsys: CapSys) -> None:
    """
    This is not the same as test_token_type_invalid() since 
    this checks our integration with pydoctor and validates we **actually** trigger 
    the warnings.
    """
    # from napoleon upstream:
    # unbalanced parenthesis in type expression
    # unbalanced square braces in type expression
    # invalid value set (missing closing brace)
    # invalid value set (missing opening brace)
    # malformed string literal (missing closing quote)
    # malformed string literal (missing opening quote)
    # from our custom napoleon:
    # invalid type: '{before_colon}'. Probably missing colon.
    # from our integration with docutils:
    # Unexpected element in type specification field

    src = '''
    __docformat__ = 'google'
    def foo(**args):
        """
        Keyword Args:
            a (list(str): thing
            b (liststr]): stuff
            c ({1,2,3): num
            d ('1',2,3}): num or str
            e (str, '1', '2): str
            f (str, "1", 2"): str
            docformat
                Can be one of:
                - "numpy"
                - "google"
            h: things
            k: stuff
        
        :type h: stuff

            >>> python
        
        :type k: a paragraph

            another one
        """
    '''

    mod = fromText(src, modname='warns')    
    docstring2html(mod.contents['foo'])

    # Filter docstring linker warnings
    lines = [line for line in capsys.readouterr().out.splitlines() if 'Cannot find link target' not in line]
    
    # Line numbers are off because they are based on the reStructuredText version of the docstring
    # which includes much more lines because of the :type arg: fields. 
    assert '\n'.join(lines) == '''\
warns:13: bad docstring: invalid type: 'docformatCan be one of'. Probably missing colon.
warns:6: bad docstring: unbalanced parenthesis in type expression
warns:8: bad docstring: unbalanced square braces in type expression
warns:10: bad docstring: invalid value set (missing closing brace): {1
warns:12: bad docstring: invalid value set (missing opening brace): 3}
warns:14: bad docstring: malformed string literal (missing closing quote): '2
warns:16: bad docstring: malformed string literal (missing opening quote): 2"
warns:23: bad docstring: Unexpected element in type specification field: element 'doctest_block'. This value should only contain text or inline markup.
warns:27: bad docstring: Unexpected element in type specification field: element 'paragraph'. This value should only contain text or inline markup.'''

def test_process_types_with_consolidated_fields(capsys: CapSys) -> None:
    """
    Test for issue https://github.com/twisted/pydoctor/issues/765
    """
    src = '''
    class V:
        """
        Doc. 

        :CVariables:
            `id` : int
                Classvar doc.
        """
    '''
    system = model.System()

    system.options.processtypes = True
    system.options.docformat = 'restructuredtext'

    mod = fromText(src, modname='do_not_warn_please', system=system)
    attr = mod.contents['V'].contents['id']
    assert isinstance(attr, model.Attribute)
    html = getHTMLOfAttribute(attr, 'base')
    # Filter docstring linker warnings
    lines = [line for line in capsys.readouterr().out.splitlines() if 'Cannot find link target' not in line]
    assert not lines
    assert '<code>int</code>' in html

def test_process_types_doesnt_mess_with_warning_linenumber(capsys: CapSys) -> None:
    src = '''
    __docformat__ = 'epytext'
    class ConfigFileParser(object):
        """doc"""

        def parse(self, stream, stuff):
            """
            Parses the keys and values from a config file.

            @param stream: A config file input stream (such as an open file object).
            @type stream: (notfound, thing[)
            @param stuff: Stuff
            @type stuff: array_like, with L{np.bytes_} or L{np.str_} dtype
            """
    '''
    system = model.System()
    system.options.processtypes = True
    mod = fromText(src, system=system)
    docstring2html(mod.contents['ConfigFileParser'].contents['parse'])
    # These linenumbers, are correct.
    assert capsys.readouterr().out.splitlines() == [
        '<test>:11: bad docstring: unbalanced square braces in type expression', 
        '<test>:11: Cannot find link target for "notfound"', 
        '<test>:11: Cannot find link target for "thing"', 
        '<test>:13: Cannot find link target for "array_like"', 
        '<test>:13: Cannot find link target for "np.bytes_" (you can link to external docs with --intersphinx)', 
        '<test>:13: Cannot find link target for "np.str_" (you can link to external docs with --intersphinx)'
        ]

def test_process_types_doesnt_mess_with_warning_linenumber_rst(capsys: CapSys) -> None:
    src = '''
    __docformat__ = 'restructuredtext'
    class ConfigFileParser(object):
        """doc"""

        def parse(self, stream, stuff):
            """
            Parses the keys and values from a config file.
            
            :param stream: A config file input stream (such as an open file object).
            :type stream: (notfound, thing[)
            :param stuff: Stuff
            :type stuff: array_like, with `np.bytes_` or `np.str_` dtype
            """
    '''
    system = model.System()
    system.options.processtypes = True
    mod = fromText(src, system=system)
    html = docstring2html(mod.contents['ConfigFileParser'].contents['parse'])
    assert 'np.bytes_' in html
    # These linenumbers, are correct.
    assert capsys.readouterr().out.splitlines() == [
        '<test>:11: bad docstring: unbalanced square braces in type expression', 
        '<test>:11: Cannot find link target for "notfound"', 
        '<test>:11: Cannot find link target for "thing"', 
        '<test>:13: Cannot find link target for "array_like"', 
        '<test>:13: Cannot find link target for "np.bytes_" (you can link to external docs with --intersphinx)', 
        '<test>:13: Cannot find link target for "np.str_" (you can link to external docs with --intersphinx)'
        ]

def test_bug_attribute_type_not_found_reports_only_once(capsys:CapSys) -> None:
    src = '''
    __docformat__ = 'numpy'
    class MachAr:
        """
        Diagnosing machine parameters.

        Attributes
        ----------
        ibeta : int
            Radix in which numbers are represented.
        """
    '''

    mod = fromText(src)
    [docstring2html(o) for o in mod.system.allobjects.values()]
    assert capsys.readouterr().out.splitlines() == ['<test>:8: Cannot find link target for "int"']